#include "UChar.h"
#include "UUtil.h"
#include "udebug.h"

// static variables
float UChar::m_scale = 0.003f;
D3DXVECTOR3 vZ(0, 0, 1.0f);

void cpStr( const char* input, char **output )
{
    if( input )
    {
        UINT length = (UINT)::strlen(input) + 1;
        *output = new char[length];
        ::strcpy( *output, input );
    }
    else
    {
        *output = 0;
    }
}

HRESULT AllocMeshHierarchy::CreateFrame( LPCSTR Name, LPD3DXFRAME *ppNewFrame )
{
	// Remark: CreateFrame must return a non-null value for ppNewFrame,
	// otherwise D3DXLoadMeshHierarchyFromX will interpret it as 
	// a failed operation.

    FrameEx* frameEx = new FrameEx();
    if( Name )
    {
        cpStr( Name, &frameEx->Name );
    }
    else
    {
        cpStr("<no name>", &frameEx->Name );
    }

    frameEx->pMeshContainer = 0;
    frameEx->pFrameSibling = 0;
    frameEx->pFrameFirstChild = 0;
    D3DXMatrixIdentity( &frameEx->TransformationMatrix );
    D3DXMatrixIdentity( &frameEx->getCombindedTransform() );

    *ppNewFrame = frameEx;

    return D3D_OK;
}

HRESULT AllocMeshHierarchy::CreateMeshContainer(
    LPCSTR Name,
    const D3DXMESHDATA *pMeshData,
    const D3DXMATERIAL *pMaterials,
    const D3DXEFFECTINSTANCE *pEffectInstances,
    DWORD NumMaterials,
    const DWORD *pAdjacency,
    LPD3DXSKININFO pSkinInfo,
    LPD3DXMESHCONTAINER *ppNewMeshContainer )
{
	// Remark: CreateMeshContainer should always return a non-null value
	// for ppNewMeshContainer, even if we are not interested in the mesh 
	// (i.e., it contains no skin info), otherwise D3DXLoadMeshHierarchyFromX
	// will interpret it as  a failed operation.

    // Allocate a new D3DXMESHCONTAINER, and set its name
    D3DXMESHCONTAINER* meshContainer = new D3DXMESHCONTAINER();
    ::ZeroMemory( meshContainer, sizeof( D3DXMESHCONTAINER ) );
    if( Name )
    {
        cpStr( Name, &meshContainer->Name );
    }
    else
    {
        cpStr( "<no name>", &meshContainer->Name );
    }

    // save our created mesh container now because we might return early,
    // and we must _always_ return an allocated container.
    *ppNewMeshContainer = meshContainer;

    // only interected in meshes with skin info and regular meshes.
    // if a given mesh does not satisfy these requirements we do
    // not return an error, rather we simply skip loading anything
    // for this mesh container.
    if( pSkinInfo == 0 || pMeshData->Type != D3DXMESHTYPE_MESH )
    {
        return D3D_OK;
    }

    // copy material data, and allocate memory for texture file names.
    meshContainer->NumMaterials = NumMaterials;
    meshContainer->pMaterials = new D3DXMATERIAL[NumMaterials];
    for( DWORD i=0; i < NumMaterials; ++i )
    {
        D3DXMATERIAL* mtrls = meshContainer->pMaterials;
        mtrls[i].MatD3D = pMaterials[i].MatD3D;
        mtrls[i].MatD3D.Ambient = pMaterials[i].MatD3D.Diffuse;

        cpStr( pMaterials[i].pTextureFilename, &mtrls[i].pTextureFilename );
    }

    // ignore effect instances and adjacency info for this demo
    meshContainer->pEffects = 0;
    meshContainer->pAdjacency = 0;

    // save mesh and skininfo
    meshContainer->MeshData.Type = D3DXMESHTYPE_MESH;
    meshContainer->MeshData.pMesh = pMeshData->pMesh;
    meshContainer->pSkinInfo = pSkinInfo;
    pMeshData->pMesh->AddRef();
    pSkinInfo->AddRef();

    return D3D_OK;
}

HRESULT AllocMeshHierarchy::DestroyFrame( D3DXFRAME* pFrameToFree )
{
    DeleteArray( pFrameToFree->Name );
    Delete( pFrameToFree );

    return D3D_OK;
}

HRESULT AllocMeshHierarchy::DestroyMeshContainer( D3DXMESHCONTAINER *pMeshContainerToFree )
{
    DeleteArray( pMeshContainerToFree->Name );
    DeleteArray( pMeshContainerToFree->pAdjacency );
    DeleteArray( pMeshContainerToFree->pEffects );

    for( DWORD i = 0; i < pMeshContainerToFree->NumMaterials; i++ )
    {
        DeleteArray( pMeshContainerToFree->pMaterials[i].pTextureFilename );
    }

    ReleaseCom( pMeshContainerToFree->MeshData.pMesh );
    ReleaseCom( pMeshContainerToFree->pSkinInfo );

    Delete( pMeshContainerToFree );

    return D3D_OK;

}

UChar::UChar()
: m_pd3dDevice(0), m_skinnedMesh(0), m_effect(0), m_root(0),
m_maxVertInfluences(0), m_numBones(0), m_skinInfo(0), m_animCtrl(0), m_hTech(0),
m_hWorldViewProj(0), m_hFinalTransforms(0), m_hTex(0), m_fYaw(0), m_vPos(0,0,0), m_vDst(0,0,0), m_fSpeed(1.5f)
{
}

HRESULT UChar::initDeviceObjects( LPDIRECT3DDEVICE9 pd3dDevice,
                              const std::string& inputFilename,
                              UMap* pmap)
{
    m_pd3dDevice = pd3dDevice;

    m_map = pmap;

    AllocMeshHierarchy allocMeshHierarchy;
    HRESULT hr;
    if( FAILED( hr = D3DXLoadMeshHierarchyFromX(
        inputFilename.c_str(),
        D3DXMESH_MANAGED,
        m_pd3dDevice,
        &allocMeshHierarchy,
        0, // ignore user data
        &m_root,
        &m_animCtrl) ) )
    {
        return hr;
    }

    D3DXFRAME* f = findNodeWithMesh( m_root );
    if( f == 0 )
    {
        return E_FAIL;
    }
    D3DXMESHCONTAINER* meshContainer = f->pMeshContainer;
    m_skinInfo = meshContainer->pSkinInfo;
    m_skinInfo->AddRef();

    m_numBones = meshContainer->pSkinInfo->GetNumBones();
    m_finalTransforms.resize( m_numBones );
    m_combinedTransforms.resize( m_numBones, 0 );

    if( FAILED(hr = buildSkinnedMesh( meshContainer->MeshData.pMesh )) )
    {
        return hr;
    }

    buildCombinedTransforms();

    if( FAILED( hr = buildEffect() ) )
    {
        return hr;
    }

    if( FAILED( hr = buildMaterials( meshContainer ) ) )
    {
        return hr;
    }

    m_aniIdx = LOITER_INDEX;

    m_vPos.x = m_map->getRow() / 2.0f;
    m_vPos.z = m_map->getCol() / 2.0f;
    m_map->getHeight( m_vPos.x, &m_vPos.y, m_vPos.z );

    return S_OK;
}

D3DXFRAME* UChar::findNodeWithMesh( D3DXFRAME* frame )
{
    // in this demo we stipulate that the input .X file contains only one
    // mesh. so search for that one and only mesh.
    if( frame->pMeshContainer )
    {
        if( frame->pMeshContainer->MeshData.pMesh != 0 )
        {
            return frame;
        }
    }

    D3DXFRAME* f = 0;
    if( frame->pFrameSibling )
    {
        if( f = findNodeWithMesh( frame->pFrameSibling ) )
        {
            return f;
        }
    }

    if( frame->pFrameFirstChild )
    {
        if( f = findNodeWithMesh( frame->pFrameFirstChild ) )
        {
            return f;
        }
    }

    return 0;
}

HRESULT UChar::buildSkinnedMesh( ID3DXMesh* mesh )
{
    // the vertex format of the source mesh does not include vertex weights
    // nor bone index data, both of which are needed for vertex blending.
    // therefore, we must convert the source mesh to an 'indexed-blended-mesh'
    // which does have the necessary data

    DWORD numBoneComboEntries = 0;
    ID3DXBuffer* boneComboTable = 0;

    HRESULT hr;

    if( FAILED(hr = m_skinInfo->ConvertToIndexedBlendedMesh(
        mesh,
        D3DXMESH_MANAGED | D3DXMESH_WRITEONLY,
        MAX_NUM_BONES_SUPPORTED,
        0, // ignore adjacency in
        0, // ignore adjacency out
        0, // ignore face remap
        0, // ignore vertex remap
        &m_maxVertInfluences,
        &numBoneComboEntries,
        &boneComboTable,
        &m_skinnedMesh)) )
    {
        return hr;
    }

    // we do not need the bone table, so just release it
    ReleaseCom( boneComboTable );

    return S_OK;
}

//
// get an array of pointers to the combined frame transfromations. we
// store the pointers such that the ith pointer corresponds with the
// i-th offset matrix. thus, give the i-th bone, we can obtain its i-th
// offset matrix and i-th transfrom.
// we note that the skinInfo object converted the source mesh to a
// indexed-blended-mesh. It follows then that the bone indices for
// the vertices are relative to the skinInfo bone array. therefore
// we base everything relative to the offset-matrix array since that
// is the array the bone indices of the vertices are relative to.
//
inline void UChar::buildCombinedTransforms()
{
    for( UINT i = 0; i < m_numBones; i++ )
    {
        // find the frame that corresponds with the i-th bone offset matrix
        const char* boneName = m_skinInfo->GetBoneName( i );
        D3DXFRAME* frame = D3DXFrameFind( m_root, boneName );
        if( frame )
        {
            FrameEx* frameEx = static_cast<FrameEx*>( frame );
            m_combinedTransforms[i] = &frameEx->getCombindedTransform();
        }
    }
}

HRESULT UChar::buildEffect()
{
    ID3DXBuffer* errorBuffer = 0;
    D3DXCreateEffectFromFile(
        m_pd3dDevice,
        "VertBlendDynamic.txt",
        0,
        0,
        D3DXSHADER_DEBUG,
        0,
        &m_effect,
        &errorBuffer);
    if( errorBuffer )
    {
        UPRINT( (char*)errorBuffer->GetBufferPointer() );
        ReleaseCom( errorBuffer );
        ReleaseCom( m_effect );
        return E_FAIL;
    }
    else
    {
        m_hTech = m_effect->GetTechniqueByName( "VertexBlendingTech" );
        m_hWorldViewProj = m_effect->GetParameterByName( 0, "WorldViewProj" );
        m_hFinalTransforms = m_effect->GetParameterByName( 0, "FinalTransforms" );
        m_hTex = m_effect->GetParameterByName( 0, "Tex" );
    }

    return S_OK;
}

//
// in this demo we assume the skinned mesh has exactly one texture.
// so load and set that texture with the effect interface
//
HRESULT UChar::buildMaterials( D3DXMESHCONTAINER* meshContainer )
{
    D3DXMATERIAL* mtrls = meshContainer->pMaterials;
    DWORD numMtrls = meshContainer->NumMaterials;

    if( mtrls[0].pTextureFilename )
    {
        IDirect3DTexture9* tex = 0;
        HRESULT hr;
        if( FAILED( hr = D3DXCreateTextureFromFile( m_pd3dDevice, mtrls[0].pTextureFilename, &tex) ) )
        {
            return hr;
        }
        if( FAILED( hr = m_effect->SetTexture( m_hTex, tex ) ) )
        {
            return hr;
        }
        ReleaseCom(tex);
    }
    return S_OK;
}

//
// release resources
//
void UChar::release()
{
    if( m_root )
    {
        AllocMeshHierarchy allocMeshHierarchy;
        D3DXFrameDestroy( m_root, &allocMeshHierarchy );
        m_root = 0;
    }

    ReleaseCom( m_skinnedMesh );
    ReleaseCom( m_skinInfo );
    ReleaseCom( m_animCtrl );
    ReleaseCom( m_effect );
}

//
// animate the mesh, the AnimationController has pointers to the
// hierarchy frame transform matrices, the AnimationController updates
// these matrices to reflect the give pose at the current time by
// interpolating between animation keyframes
//
void UChar::frameMove( float deltaTime, D3DXMATRIX& worldViewProj )
{
    // ĳ ̵
    if( m_aniIdx == WALK_INDEX )
    {
        // ̷ 
        D3DXVECTOR3 vDelta = m_vDir * m_fSpeed * deltaTime;
        D3DXVECTOR3 vMax = m_vDst - m_vPos;
        if( D3DXVec3Length( &vMax ) < D3DXVec3Length( &vDelta ) )
        {
            // ǥ 
            vDelta = vMax;
            m_aniIdx = LOITER_INDEX;
        }

        m_vPos = m_vPos + vDelta;
        //  ġ  
        m_map->getHeight( m_vPos.x, &m_vPos.y, m_vPos.z );
        //UPRINT("(%f,%f,%f) (%f,%f,%f)", m_vDst.x, m_vDst.y, m_vDst.z, m_vPos.x, m_vPos.y, m_vPos.z );
    }

    // ĳ ִϸ̼  
    play( m_aniIdx );

    m_animCtrl->AdvanceTime( deltaTime, 0 );

    // now that the frames are updated to the current pose, recurse
    // down the tree and generate a frame's combined-transform
    D3DXMATRIX matR, matY, matS;
    // tiny_4anim.x z   Ǿ Ƿ X  -90 ȸŲ.
    D3DXMatrixRotationX( &matR, -D3DX_PI/2 );
    // ĳ  
    D3DXMatrixRotationY( &matY, m_fYaw );
    // ũ⸦ Ѵ.
    D3DXMatrixScaling( &matS, m_scale, m_scale, m_scale );
    combineTransforms( static_cast<FrameEx*>(m_root), matS * matR * matY );

    // add the offset-transform, note that this premultiplies
    // the combined-transform
    D3DXMATRIX offsetTemp, combinedTemp;
    for( UINT i = 0; i<m_numBones; i++ )
    {
        offsetTemp = *m_skinInfo->GetBoneOffsetMatrix(i);
        combinedTemp = *m_combinedTransforms[i];
        m_finalTransforms[i] = offsetTemp * combinedTemp;
    }

    // ĳ ̵
    D3DXMATRIX matMWVP;
    D3DXMatrixTranslation( &matMWVP, m_vPos.x, m_vPos.y, m_vPos.z );

    matMWVP = matMWVP * worldViewProj;

    // finally, set the final-transform matrix array to the effect
    m_effect->SetMatrix( m_hWorldViewProj, &matMWVP );
    m_effect->SetMatrixArray( m_hFinalTransforms, &m_finalTransforms[0], m_finalTransforms.size() );
}

//
// recurse down the tree, combining transforms as we go along.
// any node N's combinded transform C is determinned by its local
// transform L, and the combined transform P of it's parent.
// thus, C = L * P
//
void UChar::combineTransforms( FrameEx* frame, D3DXMATRIX &P )
{
    // save some references to economize line space.
    D3DXMATRIX& L = frame->TransformationMatrix;
    D3DXMATRIX& C = frame->getCombindedTransform();

    C = L * P;

    FrameEx* sibling = (FrameEx*)frame->pFrameSibling;
    FrameEx* firstChild = (FrameEx*)frame->pFrameFirstChild;

    // recurse down siblings
    if( sibling )
    {
        combineTransforms( sibling, P );
    }

    // Recurse to first child
    if( firstChild )
    {
        combineTransforms( firstChild, C );
    }
}

//
// play animation
//
void UChar::play( ANIMATIONSET_INDEX idx )
{
  	ID3DXAnimationSet* aniSet = 0;
	m_animCtrl->GetAnimationSet(idx, &aniSet);
	m_animCtrl->SetTrackAnimationSet(0, aniSet);
	m_animCtrl->ResetTime();
}

//
// render a scene
//
void UChar::render()
{
    m_effect->SetTechnique( m_hTech );
    UINT numPasses = 0;
    m_effect->Begin( &numPasses, 0 );

    for( UINT i = 0; i<numPasses; i++ )
    {
        m_effect->BeginPass(i);

        // draw the one and only subset
        m_skinnedMesh->DrawSubset(0);

        m_effect->EndPass();
    }
    m_effect->End();
}

//
// make this character move to the destination
//
void UChar::move( D3DXVECTOR3 &dst )
{
    // ִϸ̼  
    m_aniIdx = WALK_INDEX;

    // ǥ  
    m_vDst = dst;
    //m_vDst.y = 0;
    
    // ĳ ̵  
    m_vDir = m_vDst - m_vPos;
    // X-Z 鿡 projection
    m_vDir.y = 0;
    // normalize
    D3DXVec3Normalize( &m_vDir, &m_vDir );

    // z+࿡  
    float costheta = D3DXVec3Dot( &vZ, &m_vDir );
    m_fYaw = acosf( costheta );

    if( m_vDir.x < 0.0f )
    {
        m_fYaw = -m_fYaw;
    }

    //UPRINT( "vDir(%f,%f,%f) angle(%f)", vDir.x, vDir.y, vDir.z, m_fYaw/D3DX_PI*180 );
}